-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THEventManager = {}
local THEventManager_mt = THUtils.createClass(THEventManager, Object)
InitObjectClass(THEventManager, "THEventManager")
THEventManagerData = {}
local THEventManagerData_mt = THUtils.createClass(THEventManagerData)
local function getCustomData(self)
    local customData = nil
    if self ~= nil then
        customData = rawget(self, self)
    end
    return customData
end
THEventManager.getCustomData = getCustomData
function THEventManager.new(target, customMt)
    customMt = customMt or THEventManager_mt
    if THUtils.argIsValid(THUtils.getIsType(target, THCore), "target", target)
        and THUtils.argIsValid(type(customMt) == "table", "customMt", customMt)
    then
        local self = Object.new(target.isServer, target.isClient, customMt)
        if self ~= nil then
            local customData = THEventManagerData.newData(self, target)
            if customData ~= nil then
                rawset(self, self, customData)
                return self
            end
            self:delete()
        end
    end
end
function THEventManager:update(dt, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local customData = getCustomData(self)
            if customData ~= nil and customData.isEnabled then
                local target = customData.target
                if target.onUpdate ~= nil then
                    target:onUpdate(dt)
                end
                self:raiseEvent("onUpdate", dt)
            end
        end)
        return ...
    end
    return appendFunc(THEventManager:superClass().update(self, dt, ...))
end
function THEventManager:updateTick(dt, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local customData = getCustomData(self)
            if customData ~= nil and customData.isEnabled then
                local target = customData.target
                if target.onUpdateTick ~= nil then
                    target:onUpdateTick(dt)
                end
                self:raiseEvent("onUpdateTick", dt)
            end
        end)
        return ...
    end
    return appendFunc(THEventManager:superClass().updateTick(self, dt, ...))
end
function THEventManager:delete(...)
    local function appendFunc(...)
        THUtils.call(function()
            local customData = getCustomData(self)
            if customData ~= nil then
                customData.isEnabled = false
            end
        end)
        return ...
    end
    return appendFunc(THEventManager:superClass().delete(self, ...))
end
function THEventManager:readStream(streamId, connection, ...)
    local function appendFunc(...)
        local customData = getCustomData(self)
        if customData ~= nil then
            if streamReadBool(streamId) then -- server is enabled
                local target = customData.target
                if target.onReadStream ~= nil then
                    target:onReadStream(streamId, connection)
                end
                self:raiseEvent("onReadStream", streamId, connection)
            else
                customData.isEnabled = false
            end
        end
        return ...
    end
    return appendFunc(THEventManager:superClass().readStream(self, streamId, connection, ...))
end
function THEventManager:writeStream(streamId, connection, ...)
    local function appendFunc(...)
        local customData = getCustomData(self)
        if customData ~= nil then
            if streamWriteBool(streamId, customData.isEnabled == true) then
                local target = customData.target
                if target.onWriteStream ~= nil then
                    target:onWriteStream(streamId, connection)
                end
                self:raiseEvent("onWriteStream", streamId, connection)
            end
        end
        return ...
    end
    return appendFunc(THEventManager:superClass().writeStream(self, streamId, connection, ...))
end
function THEventManager:readUpdateStream(streamId, timeStamp, connection, ...)
    local function appendFunc(...)
        local customData = getCustomData(self)
        if customData ~= nil then
            if streamReadBool(streamId) then -- server is enabled
                local target = customData.target
                if target.onReadUpdateStream ~= nil then
                    target:onReadUpdateStream(streamId, timeStamp, connection)
                end
                self:raiseEvent("onReadUpdateStream", streamId, timeStamp, connection)
            end
        end
        return ...
    end
    return appendFunc(THEventManager:superClass().readUpdateStream(self, streamId, timeStamp, connection, ...))
end
function THEventManager:writeUpdateStream(streamId, connection, dirtyMask, ...)
    local function appendFunc(...)
        local customData = getCustomData(self)
        if customData ~= nil then
            if streamWriteBool(streamId, customData.isEnabled == true) then
                local target = customData.target
                if target.onWriteUpdateStream ~= nil then
                    target:onWriteUpdateStream(streamId, connection, dirtyMask)
                end
                self:raiseEvent("onWriteUpdateStream", streamId, connection, dirtyMask)
            end
        end
        return ...
    end
    return appendFunc(THEventManager:superClass().writeUpdateStream(self, streamId, connection, dirtyMask, ...))
end
function THEventManager:raiseActive(...)
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled then
        return THEventManager:superClass().raiseActive(self, ...)
    end
end
function THEventManager:raiseDirtyFlags(dirtyFlag, ...)
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled then
        return THEventManager:superClass().raiseDirtyFlags(self, dirtyFlag, ...)
    end
end
function THEventManager:addEventListener(listener)
    if THUtils.argIsValid(type(listener) == "table", "listener", listener) then
        local customData = getCustomData(self)
        if customData ~= nil and customData.isEnabled then
            if listener == self then
                THUtils.warningMsg("Cannot add event manager target to listeners table")
                return
            end
            if customData.target ~= nil and listener == customData.target then
                THUtils.warningMsg("Cannot add target environment to listeners table")
                return
            end
            local isListenerFound = false
            for _, otherListener in pairs(customData.eventListeners) do
                if listener == otherListener then
                    isListenerFound = true
                    break
                end
            end
            if not isListenerFound then
                table.insert(customData.eventListeners, listener)
            end
        end
    end
end
function THEventManager:removeEventListener(listener)
    local customData = getCustomData(self)
    if customData ~= nil then
        local numListeners = #customData.eventListeners
        if listener ~= nil and numListeners > 0 then
            local listenerIndex = 1
            while true do
                if customData.eventListeners[listenerIndex] == nil then
                    break
                end
                if customData.eventListeners[listenerIndex] == listener then
                    table.remove(customData.eventListeners, listenerIndex)
                else
                    listenerIndex = listenerIndex + 1
                end
            end
        end
    end
end
function THEventManager:raiseEvent(eventName, ...)
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled then
        local numListeners = #customData.eventListeners
        if numListeners > 0 then
            for listenerIndex = 1, numListeners do
                local listener = customData.eventListeners[listenerIndex]
                if listener[eventName] ~= nil then
                    listener[eventName](listener, ...)
                end
            end
        end
    end
end
function THEventManagerData.newData(parent, target)
    if THUtils.argIsValid(THUtils.getIsType(parent, THEventManager), "parent", parent)
        and THUtils.argIsValid(THUtils.getIsType(target, THCore), "target", target)
    then
        local self = setmetatable({}, THEventManagerData_mt)
        self.parent = parent
        self.target = target
        self.eventListeners = {}
        self.isEnabled = true
        THUtils.setFunctionHook("XMLManager", "createSchemas", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("XMLManager", "initSchemas", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("Player", "readStream", THEventManagerData, "hook_initialReadStream", false, self)
        THUtils.setFunctionHook("Player", "writeStream", THEventManagerData, "hook_initialWriteStream", false, self)
        THUtils.setFunctionHook("Server", "sendEventIds", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("Vehicle", "init", THEventManagerData, "hook_vehicleInit", false, self)
        THUtils.setFunctionHook("HUD", "createDisplayComponents", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("Mission00", "setMissionInfo", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("Mission00", "loadEnvironment", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("FSBaseMission", "loadMapFinished", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("BaseMission", "registerActionEvents", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("BaseMission", "unregisterActionEvents", THEventManagerData, nil, false, self)
        THUtils.setFunctionHook("FSCareerMissionInfo", "saveToXMLFile", THEventManagerData, nil, false, self)
        return self
    end
end
function THEventManagerData.hook_createSchemas(superFunc, self, ...)
    THUtils.call(function()
        local parent = self.parent
        local target = self.target
        if not target.isPreInitFinished then
            if target.onPreInit ~= nil then
                target:onPreInit()
            end
            parent:raiseEvent("onPreInit")
            target.isPreInitFinished = true
        end
    end)
    return superFunc(...)
end
function THEventManagerData.hook_initSchemas(superFunc, self, ...)
    local function appendFunc(...)
        local parent = self.parent
        local target = self.target
        target:addProtectedSubTask(function()
            if not target.isPostInitFinished then
                if target.onPostInit ~= nil then
                    target:onPostInit()
                end
                parent:raiseEvent("onPostInit")
                target.isPostInitFinished = true
            end
        end)
        return ...
    end
    return appendFunc(superFunc(...))
end
function THEventManagerData.hook_vehicleInit(superFunc, self, ...)
    THUtils.call(function()
        local parent = self.parent
        local target = self.target
        target:runDelayedLoadTasks()
        parent:raiseEvent("runDelayedLoadTasks")
    end)
    return superFunc(...)
end
function THEventManagerData.hook_createDisplayComponents(superFunc, self, hud, uiScale, ...)
    local target = self.target
    local parent = self.parent
    THUtils.call(function()
        target:onPreCreateHUDComponents(hud, uiScale)
        parent:raiseEvent("onPreCreateHUDComponents", self, uiScale)
    end)
    local function appendFunc(...)
        THUtils.call(function()
            target:onHUDComponentsCreated(hud, uiScale)
            parent:raiseEvent("onHUDComponentsCreated", hud, uiScale)
        end)
        return ...
    end
    return appendFunc(superFunc(hud, uiScale, ...))
end
function THEventManagerData.hook_setMissionInfo(superFunc, self, mission, missionInfo, missionDynamicInfo, ...)
    local function appendFunc(...)
        local parent = self.parent
        local target = self.target
        target:addProtectedTask(function()
            if target.onSetMissionInfo ~= nil then
                target:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
            end
            parent:raiseEvent("onSetMissionInfo", mission, missionInfo, missionDynamicInfo)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, missionInfo, missionDynamicInfo, ...))
end
function THEventManagerData.hook_loadEnvironment(superFunc, self, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local parent = self.parent
            local target = self.target
            if target.onLoadEnvironment ~= nil then
                target:onLoadEnvironment(mission.environment, mission)
            end
            parent:raiseEvent("onLoadEnvironment", mission.environment, mission)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THEventManagerData.hook_sendEventIds(superFunc, self, server, connection, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local parent = self.parent
            local target = self.target
            if target.onSendEventIds ~= nil then
                target:onSendEventIds(connection)
            end
            parent:raiseEvent("onSendEventIds", connection)
        end)
        return ...
    end
    return appendFunc(superFunc(server, connection, ...))
end
function THEventManagerData.hook_initialReadStream(superFunc, self, player, streamId, connection, ...)
    local function appendFunc(...)
        if connection:getIsServer() then -- client reading from server
            local parent = self.parent
            local updaterId = NetworkUtil.readNodeObjectId(streamId)
            parent:readStream(streamId, connection)
            g_client:finishRegisterObject(parent, updaterId)
        end
        return ...
    end
    return appendFunc(superFunc(player, streamId, connection, ...))
end
function THEventManagerData.hook_initialWriteStream(superFunc, self, player, streamId, connection, ...)
    local function appendFunc(...)
        if not connection:getIsServer() then -- server writing to client
            local parent = self.parent
            local updaterId = NetworkUtil.getObjectId(parent)
            NetworkUtil.writeNodeObjectId(streamId, updaterId)
            parent:writeStream(streamId, connection)
            g_server:registerObjectInStream(connection, parent)
        end
        return ...
    end
    return appendFunc(superFunc(player, streamId, connection, ...))
end
function THEventManagerData.hook_loadMapFinished(superFunc, self, mission, ...)
    local function appendFunc(...)
        if not mission.cancelLoading then
            THUtils.call(function()
                local parent = self.parent
                local target = self.target
                if target.onLoadMapFinished ~= nil then
                    target:onLoadMapFinished(mission)
                end
                parent:raiseEvent("onLoadMapFinished", mission)
            end)
        end
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THEventManagerData.hook_registerActionEvents(superFunc, self, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local parent = self.parent
            local target = self.target
            if target.onRegisterActionEvents ~= nil then
                target:onRegisterActionEvents(mission)
            end
            parent:raiseEvent("onRegisterActionEvents", mission)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THEventManagerData.hook_unregisterActionEvents(superFunc, self, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local parent = self.parent
            local target = self.target
            if target.onUnregisterActionEvents ~= nil then
                target:onUnregisterActionEvents(mission)
            end
            parent:raiseEvent("onUnregisterActionEvents", mission)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THEventManagerData.hook_saveToXMLFile(superFunc, self, missionInfo, ...)
    local function appendFunc(...)
        local parent = self.parent
        local target = self.target
        if parent.isServer and target.savegame ~= nil then
            THUtils.call(function()
                if target.saveToSavegame ~= nil then
                    target:saveToSavegame()
                end
                parent:raiseEvent("saveToSavegame")
            end)
        end
        return ...
    end
    return appendFunc(superFunc(missionInfo, ...))
end