-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THSpecializationManager = {}
local THSpecializationManager_mt = THUtils.createClass(THSpecializationManager, THAbstractSystem)
THSpecializationManager.SPEC_TYPE = {
    VEHICLE = 1,
    PLACEABLE = 2
}
THSpecializationManager.SPEC_TYPE_ID = {}
for typeId, typeIndex in pairs(THSpecializationManager.SPEC_TYPE) do
    THSpecializationManager.SPEC_TYPE_ID[typeIndex] = typeId
end
function THSpecializationManager.new(owner, dataKey, modName, customMt)
    customMt = customMt or THSpecializationManager_mt
    local self = THAbstractSystem.new(owner, dataKey, modName, customMt)
    if self ~= nil then
        local globalEnv = THUtils.getGlobalEnvironment()
        self.specsByType = {}
        for _, specTypeIndex in pairs(THSpecializationManager.SPEC_TYPE) do
            self.specsByType[specTypeIndex] = {}
        end
        self.specManagers = {}
        self.specManagers[self.SPEC_TYPE.VEHICLE] = globalEnv.g_specializationManager
        self.specManagers[self.SPEC_TYPE.PLACEABLE] = globalEnv.g_placeableSpecializationManager
        THUtils.setFunctionHook("TypeManager", "finalizeTypes", THSpecializationManager, nil, false, self)
    end
    return self
end
function THSpecializationManager:getSpecTypeManager(specType)
    local targetType = type(specType)
    if targetType == "number" then
        return self.specManagers[specType]
    elseif targetType == "string" then
        local specTypeIndex = THSpecializationManager.SPEC_TYPE[specType:upper()]
        if specTypeIndex ~= nil then
            return self.specManagers[specTypeIndex]
        end
    end
end
function THSpecializationManager:getSpecListByType(specType)
    local targetType = type(specType)
    local specTypeIndex = 0
    if targetType == "number" then
        specTypeIndex = specType
    elseif targetType == "string" then
        specTypeIndex = THSpecializationManager.SPEC_TYPE[specType:upper()]
    else
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "specType", specType)
    end
    if specTypeIndex ~= nil and specTypeIndex > 0 then
        local specsArray = self.specsByType[specTypeIndex]
        if specsArray ~= nil then
            return specsArray, #specsArray
        end
    end
    return nil, 0
end
function THSpecializationManager:addSpecialization(specType, specName, className, modName, filename)
    local specManager = self:getSpecTypeManager(specType)
    local thModManager = self.mainEnvironment.thModManager
    if specManager == nil then
        THUtils.stackMsg("No specialization manager for specType (%s)", specType)
    else
        local modData = thModManager:getLoadedMod(modName)
        local specTypeId = specManager.typeName:upper()
        local specTypeIndex = THSpecializationManager.SPEC_TYPE[specTypeId]
        if modData == nil then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "modName", modName)
        else
            local fullSpecName = string.format("%s.%s", modData.name, specName)
            local fullClassName = string.format("%s.%s", modData.name, className)
            local fullFilename = Utils.getFilename(filename, modData.path)
            if specManager:getSpecializationByName(fullSpecName) == nil then
                specManager:addSpecialization(fullSpecName, fullClassName, fullFilename, modData.name)
                local specInfo = specManager:getSpecializationByName(fullSpecName)
                if specInfo == nil then
                    THUtils.stackMsg("Failed to add %s specialization (%s)", specManager.typeName, specName)
                else
                    table.insert(self.specsByType[specTypeIndex], specInfo)
                    THUtils.displayMsg("Added %s specialization: %s", specManager.typeName, specInfo.name)
                end
            end
        end
    end
end
function THSpecializationManager.hook_finalizeTypes(superFunc, self, typeManager, ...)
    THUtils.call(function()
        local objectTypeList = typeManager:getTypes()
        local specsArray, numSpecs = self:getSpecListByType(typeManager.typeName)
        if numSpecs > 0 then
            for _, specInfo in ipairs(specsArray) do
                local specName = specInfo.name
                local specClass = THUtils.getClassObject(specInfo.className)
                local numTimesAdded = 0
                for objectTypeName, objectType in pairs(objectTypeList) do
                    if objectType.specializationsByName[specName] == nil
                        and specClass.prerequisitesPresent(objectType.specializations)
                    then
                        typeManager:addSpecialization(objectTypeName, specName)
                        numTimesAdded = numTimesAdded + 1
                    end
                end
                if numTimesAdded > 0 then
                    THUtils.displayMsg("Added %s specialization %q to %d objects", typeManager.typeName, specName, numTimesAdded)
                end
            end
        end
    end)
    return superFunc(typeManager, ...)
end