-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THUtils = {}
THMessage = {
    INTERNAL_ERROR = "An internal error has occurred",
    ARGUMENT_INVALID = "Invalid argument %q (%s) in function call",
    MAIN_ENVIRONMENT_REQUIRED = "Parent environment must be a main environment",
    INITIALIZATION_FAILED = "Failed to initialize %s (%s) which could cause errors",
    FILE_NOT_FOUND = "Could not find file: %s",
    LOADING = "Loading: %s",
    COMMAND_REQUIRES_MASTER_USER = "Console command requires admin rights to send"
}
THXMLMessage = {
    DUPLICATE_ENTRY = "duplicate entry found",
    INVALID_VALUE = "missing or invalid value",
    INVALID_FILL_TYPE = "missing or invalid fillType",
    INVALID_FRUIT_TYPE = "missing or invalid fruitType",
    VALUE_OUT_OF_RANGE = "must be between %s and %s",
    VALUE_GREATER_ZERO = "must be greater than zero",
    VALUE_GREATER_EQUAL_ZERO = "must be greater than or equal to zero",
    VALUE_LESS_EQUAL_AMOUNT = "must be less than or equal to %s",
    FILE_NOT_FOUND = "file not found"
}
local thModName = g_currentModName
function THUtils.displayMsg(text, ...)
    if text == "" then
        print(text)
    else
        if text == nil or type(text) ~= "string" then
            text = string.format("ERROR: Invalid displayMsg text, %q", text)
        else
            if select("#", ...) > 0 then
                text = string.format(text, ...)
            end
        end
        local msgString = string.format("** [%s]: %s", thModName, text)
        print(msgString)
    end
end
function THUtils.warningMsg(text, ...)
    if type(text) == "string" then
        text = "WARNING: " .. text
    end
    THUtils.displayMsg(text, ...)
end
function THUtils.errorMsg(text, ...)
    if type(text) == "string" then
        text = "ERROR: " .. text
    end
    THUtils.displayMsg(text, ...)
end
function THUtils.stackMsg(text, ...)
    THUtils.errorMsg(text, ...)
    printCallstack()
end
function THUtils.assert(expression, text, ...)
    if not expression then
        THUtils.stackMsg(text, ...)
    end
    return expression
end
function THUtils.msgOnTrue(expression, text, ...)
    if expression then
        THUtils.stackMsg(text, ...)
        return true
    end
    return false
end
function THUtils.argIsValid(expression, argName, argValue)
    if not expression then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, argName, argValue)
        return false
    end
    return true
end
function THUtils.xmlLoadMsg(xmlKey, attrName, attrValue, isError, text, ...)
    local xmlText = "XML Warning: %s"
    if isError == true then
        xmlText = "XML Error: %s"
    end
    THUtils.displayMsg(xmlText, xmlKey)
    if attrName == nil then
        THUtils.displayMsg(">> " .. text, ...)
    else
        local newText = string.format("Attribute (%s=%q) ", attrName, attrValue)
        THUtils.displayMsg(">> " .. newText .. text, ...)
    end
end
function THUtils.printTable(target, ...)
    local targetType = type(target)
    if target == nil or (targetType == "table" and next(target) == nil) then
        THUtils.displayMsg("Table (%s) is empty", target)
    elseif targetType ~= "table" then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "target", target)
    elseif g_thDebugTools ~= nil then
        g_thDebugTools:printTable(target, ...)
    else
        THUtils.displayMsg("Printing %s", target)
        for k, v in pairs(target) do
            THUtils.displayMsg("[%s] = %s", k, v)
        end
        THUtils.displayMsg("")
    end
end
THUtils.dumpTable = THUtils.printTable
function THUtils.getIsMasterUser()
    if g_currentMission ~= nil then
        return g_currentMission.isMasterUser == true
    end
    return false
end
function THUtils.getGlobalEnvironment()
    local mt = getmetatable(_G)
    if mt ~= nil then
        if type(mt.__index) == "table" and mt.__index._G ~= nil then
            return mt.__index._G
        end
    end
    THUtils.warningMsg("Could not find global environment")
    return _G
end
function THUtils.getModEnvironment(modName)
    if modName == nil then
        return _G
    else
        local globalEnv = THUtils.getGlobalEnvironment()
        local modEnv = globalEnv[modName]
        if modEnv ~= nil then
            if modEnv._G ~= nil then
                return modEnv._G
            end
            THUtils.warningMsg("Could not find mod environment (%s)", modName)
        end
        return modEnv
    end
end
function THUtils.getClassObject(className, modName)
    local globalEnv = THUtils.getGlobalEnvironment()
    if modName ~= nil then
        className = string.format("%s.%s", modName, className)
    end
    return globalEnv.ClassUtil.getClassObject(className)
end
function THUtils.pack(...)
    return { n = select("#", ...), ... }
end
function THUtils.unpack(target, index, endIndex)
    if endIndex == nil and target.n ~= nil then
        index = THUtils.getNoNil(index, 1)
        endIndex = target.n
    end
    return unpack(target, index, endIndex)
end
function THUtils.getNoNil(target, replacement, includeEmptyStrings)
    if target == nil or (target == "" and includeEmptyStrings == true) then
        return replacement
    end
    return target
end
function THUtils.pcall(func, pa,pb,pc,pd,pe,pf,pg,ph,pi,pj,pk,pl,pm,pn,po,pp,pq,pr,ps,pt)
    local function errHandler(errMsg)
        errMsg = THUtils.getNoNil(errMsg, THMessage.INTERNAL_ERROR)
        if type(errMsg) == "string" then
            THUtils.stackMsg(errMsg)
        else
            printCallstack()
        end
    end
    local function protectedFunc()
        return func(pa,pb,pc,pd,pe,pf,pg,ph,pi,pj,pk,pl,pm,pn,po,pp,pq,pr,ps,pt)
    end
    return xpcall(protectedFunc, errHandler)
end
function THUtils.call(func, ...)
    local function appendFunc(rSuccess, ret1, ...)
        if rSuccess == true then
            return ret1, ...
        end
    end
    return appendFunc(THUtils.pcall(func, ...))
end
function THUtils.toBoolean(value)
    if value == true or value == false then
        return value
    elseif type(value) == "string" then
        local upperVal = value:upper()
        if upperVal == "TRUE" then
            return true
        elseif upperVal == "FALSE" then
            return false
        end
    end
end
function THUtils.getIsType(value, typeEntry)
    local valType = type(value)
    if valType == typeEntry then -- exact match (fastest)
        return true
    end
    local compareType = type(typeEntry)
    if compareType == "table" then
        if valType == "table" and value.isa ~= nil and value:isa(typeEntry) then
            return true
        end
    elseif compareType == "string" then
        if string.find(typeEntry, "^" .. valType .. " ")
            or string.find(typeEntry, " " .. valType .. "$")
            or string.find(typeEntry, " " .. valType .. " ")
        then
            return true
        end
    else
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "typeEntry", typeEntry)
    end
    return false
end
function THUtils.floor(value, precision)
    precision = precision or 0
    local factor = 10 ^ precision
    return math.floor(value * factor) / factor
end
function THUtils.ceil(value, precision)
    precision = precision or 0
    local factor = 10 ^ precision
    return math.ceil(value * factor) / factor
end
function THUtils.round(value, precision)
    precision = precision or 0
    local factor = 10 ^ precision
    return math.floor((value * factor) + 0.5) / factor
end
function THUtils.formatPercent(value, precision)
    precision = tonumber(precision) or 0
    precision = math.floor(precision)
    local dispValue = THUtils.round(value * 100, precision)
    local textValue = string.format("%0." .. tostring(precision) .. "f%%", dispValue)
    return textValue
end
function THUtils.getVector2Distance(sx,sz, ex,ez)
    return math.sqrt(((ex - sx) ^ 2) + ((ez - sz) ^ 2))
end
function THUtils.getVector3Distance(sx,sy,sz, ex,ey,ez)
    return math.sqrt(((ex - sx) ^ 2) + ((ey - sy) ^ 2) + ((ez - sz) ^ 2))
end
function THUtils.getVector2AreaByLine(sx, sz, ex, ez, width)
    width = math.max(0, width or 0)
    local length = THUtils.getVector2Distance(sx, sz, ex, ez)
    local mx = (sx + ex) / 2
    local mz = (sz + ez) / 2
    if length == 0 then -- single point
        if width == 0 then -- return single point
            return mx, mz, mx, mz, mx, mz
        end
        local halfWidth = width / 2
        return mx - halfWidth, mz - halfWidth, mx + halfWidth, mz - halfWidth, mx - halfWidth, mz + halfWidth
    end
    local factor = 1 + (width / length)
    local dx = ((ex - sx) / 2) * factor
    local dz = ((ez - sz) / 2) * factor
    sx = mx - dx
    sz = mz - dz
    ex = mx + dx
    ez = mz + dz
    local rdx = dz
    local rdz = -dx
    return sx + rdx, sz + rdz, sx - rdx, sz - rdz, ex + rdx, ez + rdz
end
function THUtils.getFactorInRange(val1, val2, testVal, noClamp)
    local factor = (testVal - val1) / (val2 - val1)
    if noClamp ~= true then
        factor = MathUtil.clamp(factor, 0, 1)
    end
    return factor
end
function THUtils.getValueInRange(val1, val2, factor, isClamped)
    local newVal = val1 + (val2 - val1) * factor
    if isClamped ~= false then
        local minVal = math.min(val1, val2)
        local maxVal = math.max(val1, val2)
        newVal = MathUtil.clamp(newVal, minVal, maxVal)
    end
    return newVal
end
function THUtils.splitString(text, separator)
    local splitTable = string.split(text, separator)
    if splitTable == nil then
        splitTable = {}
    end
    return splitTable, #splitTable
end
function THUtils.properCase(text, removeSpaces)
    if text == nil then -- skip (pass through)
        return
    end
    local retText = text
    retText = string.gsub(retText, "^%l", string.upper)
    retText = string.gsub(retText, " %l", string.upper)
    if removeSpaces == true then
        retText = string.gsub(retText, " ", "")
    end
    return retText
end
function THUtils.snakeCase(text, toUpper)
    local newText = string.gsub(text, " ", "_")
    newText = string.gsub(newText, "(%l)(%u)", "%1_%2")
    if toUpper == true then
        newText = newText:upper()
    else
        newText = newText:lower()
    end
    return newText
end
function THUtils.createClass(mainClass, baseClass, isProtected, ...)
    local mt = Class(mainClass, baseClass, ...)
    if mt ~= nil and isProtected == false then
        mt.__metatable = nil
    end
    return mt
end
function THUtils.clearTable(target)
    for key in pairs(target) do
        target[key] = nil
    end
    return target
end
function THUtils.copyTable(srcTable, depth)
    depth = THUtils.getNoNil(depth, 0)
    local newTable = {}
    if THUtils.argIsValid(type(srcTable) == "table", "srcTable", srcTable) then
        for key, val in pairs(srcTable) do
            if type(val) == "table" and depth > 0 then
                newTable[key] = THUtils.copyTable(val, depth - 1)
            else
                newTable[key] = val
            end
        end
        local oldMt = getmetatable(srcTable)
        if oldMt ~= nil then
            setmetatable(newTable, oldMt)
        end
    end
    return newTable
end
function THUtils.createSubTable(parent, tableKey, clear, useRaw)
    local newTable = nil -- placeholder
    if useRaw == true then
        newTable = rawget(parent, tableKey)
    else
        newTable = parent[tableKey]
    end
    if newTable ~= nil then
        if clear == true then
            THUtils.clearTable(newTable)
        end
    else
        newTable = {}
        if useRaw == true then
            rawset(parent, tableKey, newTable)
        else
            parent[tableKey] = newTable
        end
    end
    return newTable
end
function THUtils.getTableLength(target)
    local numEntries = 0
    if target ~= nil then
        for _ in pairs(target) do
            numEntries = numEntries + 1
        end
    end
    return numEntries
end
function THUtils.getTableIsArray(target)
    local isArray = target[1] ~= nil
    if isArray then
        local numEntries = 0
        for _ in pairs(target) do
            numEntries = numEntries + 1
            if target[numEntries] == nil then
                isArray = false
                break
            end
        end
    end
    return isArray
end
function THUtils.getSpecTable(object, specName, customEnv, verbose)
    local objectName = "Unknown"
    local specKey = nil
    if customEnv ~= nil then
        specKey = string.format("spec_%s.%s", customEnv, specName)
    else
        specKey = string.format("spec_%s", specName)
    end
    local specTable = nil
    if object ~= nil then
        if object.name ~= nil then
            objectName = object.name
        end
        specTable = object[specKey]
    end
    if specTable == nil and verbose ~= false then
        THUtils.stackMsg("%s >> Cannot find specialization (%s)", objectName, specKey)
    end
    return specTable
end
function THUtils.makeSelfCallback(targetFunc, newSelf)
    if newSelf == nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "newSelf", newSelf)
    else
        local callbackFunc = function(...)
            return targetFunc(newSelf, ...)
        end
        return callbackFunc
    end
end
function THUtils.getFunction(targetId, funcName, verbose, useRaw)
    local target = targetId
    local targetType = type(target)
    if targetId == nil or targetType == "string" then
        local globalEnv = THUtils.getGlobalEnvironment()
        if targetId == nil then
            target = globalEnv
        else
            target = globalEnv[targetId]
        end
        if type(target) ~= "table" then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "targetId", targetId)
            return
        end
    elseif targetType ~= "table" then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "targetId", targetId)
    end
    if THUtils.argIsValid(type(funcName) == "string", "funcName", funcName) then
        local targetFunc = nil
        if useRaw == true then
            targetFunc = rawget(target, funcName)
        else
            targetFunc = target[funcName]
        end
        local funcType = type(targetFunc)
        if targetFunc ~= nil and funcType ~= "function" then
            targetFunc = nil
            THUtils.stackMsg("Function (%s) is of type %q, function expected", funcName, funcType)
        elseif targetFunc == nil and verbose == true then
            THUtils.stackMsg("Cannot find function (%s)", funcName)
        end
        return targetFunc, target
    end
end
function THUtils.registerFunction(target, funcName, newFunc)
    if THUtils.argIsValid(type(target) == "table", "target", target)
        and THUtils.argIsValid(type(funcName) == "string", "funcName", funcName)
        and THUtils.argIsValid(type(newFunc) == "function", "newFunc", newFunc)
    then
        local oldFunc = rawget(target, funcName)
        if oldFunc ~= nil then
            THUtils.stackMsg("Function %q already registered", funcName)
            return oldFunc
        end
        local callbackFunc = function(...)
            return newFunc(...)
        end
        rawset(target, funcName, callbackFunc)
        return callbackFunc
    end
end
function THUtils.setFunctionHook(srcTableId, srcFuncName, tgtTableId, tgtFunc, useSelf, extraArg)
    useSelf = THUtils.getNoNil(useSelf, true)
    local srcFunc, srcTable = THUtils.getFunction(srcTableId, srcFuncName, true)
    if srcFunc == nil or srcTable == nil then
        return
    end
    local tgtFuncType = type(tgtFunc)
    if tgtFunc == nil or tgtFuncType == "string" then
        local tgtFuncName = tgtFunc or string.format("hook_%s", srcFuncName)
        tgtFunc = THUtils.getFunction(tgtTableId, tgtFuncName, true)
        if tgtFunc == nil then
            return
        end
    elseif tgtFuncType ~= "function" then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "tgtFunc", tgtFunc)
        return
    end
    local function callbackFunc(p1, ...)
        if useSelf == true then
            if extraArg ~= nil then
                return tgtFunc(p1, srcFunc, extraArg, ...)
            end
            return tgtFunc(p1, srcFunc, ...)
        end
        if extraArg ~= nil then
            return tgtFunc(srcFunc, extraArg, p1, ...)
        end
        return tgtFunc(srcFunc, p1, ...)
    end
    srcTable[srcFuncName] = callbackFunc
    return callbackFunc, srcTable
end
function THUtils.registerXMLPath(schema, xmlValueType, xmlPath, xmlKeys, ...)
    local xmlKeyType = type(xmlKeys)
    local xmlKeyList = nil -- placeholder
    if xmlKeyType == "string" and xmlKeys ~= "" then
        if string.find(xmlKeys, " ") then
            xmlKeyList = THUtils.splitString(xmlKeys, " ")
        else
            schema:register(xmlValueType, xmlPath .. xmlKeys, ...)
            return
        end
    elseif xmlKeyType == "table" then
        xmlKeyList = xmlKeys
    else
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "xmlKeys", xmlKeys)
        return
    end
    if xmlKeyList ~= nil and #xmlKeyList > 0 then
        for _, xmlKey in ipairs(xmlKeyList) do
            if xmlKey ~= "" then
                schema:register(xmlValueType, xmlPath .. xmlKey, ...)
            end
        end
    end
end
function THUtils.getXMLValue(xmlFile, basePath, xmlKeys, defaultValue, ...)
    local function protectedFunc(...)
        local xmlKeyType = type(xmlKeys)
        local xmlKeyList = nil -- placeholder
        if xmlKeyType == "string" then
            if string.find(xmlKeys, " ") then
                xmlKeyList = THUtils.splitString(xmlKeys, " ")
            else
                return xmlFile:getValue(basePath .. xmlKeys, defaultValue, ...)
            end
        elseif xmlKeyType == "table" then
            xmlKeyList = xmlKeys
        else
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "xmlKeys", xmlKeys)
            return
        end
        if xmlKeyList ~= nil and #xmlKeyList > 0 then
            for i, xmlKey in ipairs(xmlKeyList) do
                local xmlKeyFull = string.format("%s%s", basePath, xmlKey)
                local retValue, ra,rb,rc,rd,re,rf,rg,rh,ri,rj = xmlFile:getValue(xmlKeyFull, nil, ...)
                if retValue ~= nil then
                    if i > 1 then
                        local msgKey1 = xmlKeyFull
                        local msgKey2 = string.gsub(xmlKeyList[1], "^#", "")
                        THUtils.warningMsg("Outdated xml key: [%s] Please use %q instead", msgKey1, msgKey2)
                    end
                    return retValue, ra,rb,rc,rd,re,rf,rg,rh,ri,rj
                end
            end
        end
    end
    local function appendFunc(rSuccess, rValue, ...) -- ensures default value is always returned
        if rSuccess and rValue ~= nil then
            return rValue, ...
        else
            return defaultValue
        end
    end
    return appendFunc(THUtils.pcall(protectedFunc, ...))
end
function THUtils.loadXMLFile(xmlName, filename, xmlSchema, ...)
    if filename ~= nil and fileExists(filename) then
        local xmlFile = nil -- placeholder
        if xmlSchema ~= nil and xmlSchema ~= false then
            if xmlSchema == true then
                xmlFile = XMLFile.loadIfExists(xmlName, filename, nil, ...)
            else
                xmlFile = XMLFile.loadIfExists(xmlName, filename, xmlSchema, ...)
            end
        else
            xmlFile = loadXMLFile(xmlName, filename, ...)
            if xmlFile ~= nil and xmlFile <= 0 then
                THUtils.stackMsg("Invalid xml file (%s)", xmlName)
                return
            end
        end
        return xmlFile
    end
end
function THUtils.deleteXMLFile(xmlFile)
    if xmlFile ~= nil then
        if type(xmlFile) == "table" then
            xmlFile:delete()
        else
            delete(xmlFile)
        end
    end
end
function THUtils.getNormalizedValues(uiScale, x, y)
    local normX, normY = 0, 0
    if x ~= nil then
        normX = x * uiScale * g_aspectScaleX / g_referenceScreenWidth
    end
    if y ~= nil then
        normY = y * uiScale * g_aspectScaleY / g_referenceScreenHeight
    end
    return normX, normY
end
function THUtils.getElementByProfileName(baseElement, profileName, recursive)
    if profileName == nil then
        return
    end
    if profileName == baseElement.profile then
        return baseElement
    end
    local elementList = baseElement.elements
    if elementList and #elementList > 0 then
        for idx = 1, #elementList do
            local element = elementList[idx]
            if recursive then
                local foundElement = THUtils.getElementByProfileName(element, profileName, recursive)
                if foundElement ~= nil then
                    return foundElement
                end
            elseif profileName == element.profile then
                return element
            end
        end
    end
end