-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_currentModName
local thModPath = g_currentModDirectory
source(thModPath .. "scripts/core/THCore.lua")
THAdvancedAddon = {}
local THAdvancedAddon_mt = THUtils.createClass(THAdvancedAddon, THCore)
THAdvancedAddon.DATA_KEY = "thAdvancedAddon"
function THAdvancedAddon.new(dataKey, customMt)
    dataKey = dataKey or THAdvancedAddon.DATA_KEY
    customMt = customMt or THAdvancedAddon_mt
    local self = THCore.new(dataKey, thModName, customMt)
    if self ~= nil then
        self.modules = {
            byId = {},
            byIndex = {}
        }
        self.MODULE = {}
        self:addConsoleCommand("setDebugEnabled", "Enables/disables specified debug flag", "consoleCommandSetDebugIsEnabled")
        self:addConsoleCommand("setDebugLevel", "Sets specified debug level", "consoleCommandSetDebugLevel")
    end
    return self
end
function THAdvancedAddon:initSettings()
    return true
end
function THAdvancedAddon:onPreInit()
    if not self.isPreInitFinished then
        THAdvancedAddon:superClass().onPreInit(self)
        local thModManager = self.thModManager
        local thSpecManager = self.thSpecManager
        local thSettingsManager = self.thSettingsManager
        source(self.modPath .. "scripts/main/managers/THFillTypeManager.lua")
        source(self.modPath .. "scripts/main/managers/THFruitTypeManager.lua")
        source(self.modPath .. "scripts/main/managers/THPrecisionFarmingManager.lua")
        source(self.modPath .. "scripts/main/managers/THStorageSystem.lua")
        self.thFillTypeManager = THFillTypeManager.new(self)
        self.thFruitTypeManager = THFruitTypeManager.new(self)
        self.thPFManager = THPrecisionFarmingManager.new(self)
        if self.thFillTypeManager ~= nil
            and self.thFruitTypeManager ~= nil
            and self.thPFManager ~= nil
            and self:initSettings()
        then
            source(self.modPath .. "scripts/core/gui/THMainMenu.lua")
            source(self.modPath .. "scripts/core/gui/THSettingsMenuFrame.lua")
            local mainMenu = THMainMenu.new(nil, nil, self.messageCenter, self.i18n, self.inputManager)
            if mainMenu ~= nil then
                self.gui:loadGui(self.modPath .. "gui/mainMenu.xml", "THAdvancedAddonMenu", mainMenu)
                self.mainMenu = mainMenu.thCustomData
                if self.mainMenu == nil then
                    THUtils.stackMsg("Failed to load main menu")
                else
                    self.mainMenu:addMenuPageFromXML("gui/frameReferences.xml", "thAdvancedAddonSettings", nil, thSettingsManager)
                end
            end
            source(self.modPath .. "scripts/main/gui/THHud.lua")
            source(self.modPath .. "scripts/main/gui/THVehicleInfoDisplay.lua")
            source(self.modPath .. "scripts/main/gui/THExtendedHudBox.lua")
            thSpecManager:addSpecialization("vehicle", "vehicle", "THVehicle", self.modName, "scripts/main/vehicles/specializations/THVehicle.lua")
            thSpecManager:addSpecialization("vehicle", "cutter", "THCutter", self.modName, "scripts/main/vehicles/specializations/THCutter.lua")
            thSpecManager:addSpecialization("vehicle", "combine", "THCombine", self.modName, "scripts/main/vehicles/specializations/THCombine.lua")
            thSpecManager:addSpecialization("placeable", "placeable", "THPlaceable", self.modName, "scripts/main/placeables/specializations/THPlaceable.lua")
            THUtils.displayMsg("Loading map: %s", self.mapName)
            local modsArray, numMods = thModManager:getLoadedModList(true)
            if numMods > 0 then
                for _, modData in ipairs(modsArray) do
                    if modData.name ~= self.modName then
                        self:loadModule(modData)
                    end
                end
            end
        end
        self.isPreInitFinished = true
    end
end
function THAdvancedAddon:onRegisterActionEvents(mission)
    THAdvancedAddon:superClass().onRegisterActionEvents(self, mission)
    if self.isClient then
        if self.mainMenu ~= nil then
            local _, eventId = self.thInputManager:registerActionEvent("THAA_OPEN_MAIN_MENU", nil, self, self.actionOpenMainMenu, false, true, false, true)
            if eventId ~= nil then
                self.actionEventIds.openMainMenu = eventId
                self.thInputManager:setActionEventTextVisibility(eventId, false)
            end
        end
    end
end
function THAdvancedAddon:actionOpenMainMenu()
    if self.mainMenu ~= nil and self.mainMenu:getIsMenuValid() then
        local mission = self.mission
        if mission ~= nil and mission.isMissionStarted
            and not mission.isSynchronizingWithPlayers
        then
            self.gui:showGui("THAdvancedAddonMenu")
        end
    end
end
function THAdvancedAddon:consoleCommandSetDebugIsEnabled(flagId, isEnabled, sendToServer)
    local usageMsg = "Usage: " .. self.dataKey .. ":setDebugEnabled flagId isEnabled sendToServer"
    isEnabled = THUtils.toBoolean(isEnabled)
    sendToServer = THUtils.toBoolean(sendToServer)
    local function showUsageMessage()
        printf(usageMsg)
        printf("")
        printf("Allowed debug flagIds:")
        local debugFlagList = THDebugUtil.getDebugFlagList()
        for otherFlagId in pairs(debugFlagList) do
            printf("- %s", otherFlagId)
        end
        printf("")
    end
    if isEnabled == nil then
        showUsageMessage()
    else
        local realFlagId = THDebugUtil.getDebugFlag(flagId)
        if realFlagId == nil or realFlagId == THDebugUtil.FLAG_UNKNOWN then
            showUsageMessage()
        else
            local newIsEnabled = THDebugUtil.setIsEnabled(realFlagId, isEnabled)
            local enabledText = "enabled"
            if newIsEnabled ~= true then
                enabledText = "disabled"
            end
            if realFlagId == "" then
                THUtils.displayMsg("Debugger is %s", enabledText)
            else
                THUtils.displayMsg("Debugging for %q is %s", realFlagId, enabledText)
            end
            if not self.isServer and sendToServer == true then
                THSetIsDebugEnabledEvent.sendEvent(realFlagId, newIsEnabled)
            end
        end
    end
end
function THAdvancedAddon:consoleCommandSetDebugLevel(debugLevelId, sendToServer)
    local usageMsg = "Usage: " .. self.dataKey .. ":setDebugLevel debugLevel sendToServer"
    local debugLevel = THDebugUtil.getDebugLevelById(debugLevelId)
    local function showUsageMsg()
        printf(usageMsg)
        printf("")
        printf("Allowed debug level ids:")
        for id, level in pairs(THDebugLevel) do
            printf("[%s]: %s", id, level)
        end
        printf("")
    end
    if debugLevel == nil then
        showUsageMsg()
        return
    end
    THDebugUtil.setDebugLevel(debugLevel)
    sendToServer = THUtils.toBoolean(sendToServer)
    if not self.isServer and sendToServer == true then
        THSetDebugLevelEvent.sendEvent(debugLevel)
    end
end
function THAdvancedAddon:loadModule(modData)
    if modData.name == self.modName then
        return
    end
    local xmlKey = string.format("modDesc.%s", self.dataKey)
    local xmlFile = XMLFile.loadIfExists("THModDescXML", modData.file)
    if xmlFile ~= nil and xmlFile:hasProperty(xmlKey) then
        local modEnv = THUtils.getModEnvironment(modData.name)
        local reqVersion = xmlFile:getString(xmlKey .. "#version", "?.?.?.?")
        if reqVersion ~= self.modData.version.text then
            local wrongVersionText = self.i18n:getText("thDialog_moduleVersionMismatch")
            wrongVersionText = string.format(wrongVersionText, modData.name, reqVersion, self.modData.version.text)
            THUtils.warningMsg(wrongVersionText)
            self.gui:showInfoDialog({ text = wrongVersionText })
        elseif modEnv == nil then
            THUtils.stackMsg("Cannot find environment for mod (%s)", modData.name)
        else
            xmlFile:iterate(xmlKey .. ".modules.module", function(_, pModuleKey)
                local moduleName = xmlFile:getString(pModuleKey .. "#name", "")
                local className = xmlFile:getString(pModuleKey .. "#class", "")
                local filename = xmlFile:getString(pModuleKey .. "#filename", "")
                local isModInitialized = false
                if moduleName == "" or className == "" or filename == "" then
                    THUtils.xmlLoadMsg(pModuleKey, nil, nil, true, "invalid entry")
                else
                    local moduleId = moduleName:upper()
                    local constId = THUtils.snakeCase(moduleName, true)
                    filename = Utils.getFilename(filename, modData.path)
                    if self.modules.byId[moduleId] ~= nil
                        or self.MODULE[constId] ~= nil
                    then
                        THUtils.xmlLoadMsg(pModuleKey, "moduleName", moduleName, true, THXMLMessage.DUPLICATE_ENTRY)
                    elseif not fileExists(filename) then
                        THUtils.xmlLoadMsg(pModuleKey, "filename", filename, true, THXMLMessage.FILE_NOT_FOUND)
                    else
                        if not isModInitialized then
                            local modMeta = getmetatable(modEnv)
                            local oldMetaIndex = modMeta.__index
                            local metaIndexType = type(oldMetaIndex)
                            if metaIndexType ~= "table" and metaIndexType ~= "function" then
                                THUtils.stackMsg("Cannot find metatable for module (%s)", moduleName)
                            else
                                local function metaIndex(pTable, pKey)
                                    local newValue = _G[pKey]
                                    if newValue == nil then
                                        if metaIndexType == "function" then
                                            newValue = oldMetaIndex(pTable, pKey)
                                        elseif metaIndexType == "table" then
                                            newValue = oldMetaIndex[pKey]
                                        end
                                    end
                                    return newValue
                                end
                                modMeta.__index = metaIndex
                                isModInitialized = true
                            end
                        end
                        if isModInitialized then
                            source(filename, modData.name)
                            local moduleClass = THUtils.getClassObject(className, modData.name)
                            if moduleClass == nil then
                                THUtils.xmlLoadMsg(pModuleKey, "className", className, true, THXMLMessage.INVALID_VALUE)
                            else
                                if moduleClass.newModule ~= nil then
                                    setfenv(moduleClass.newModule, modEnv)
                                    local moduleTarget = THUtils.call(moduleClass.newModule, self, modData)
                                    if moduleTarget == nil then
                                        THUtils.stackMsg("Could not create module (%s)", moduleName)
                                    else
                                        local moduleData = self:addModule(moduleName, modData.name, moduleTarget)
                                        if moduleData ~= nil then
                                            THUtils.displayMsg("Loaded module: %s", moduleData.name)
                                        end
                                    end
                                end
                            end
                        end
                    end
                end
            end)
        end
    end
end
function THAdvancedAddon:addModule(name, modName, target)
    local thModManager = self.thModManager
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name)
        and THUtils.argIsValid(type(modName) == "string" and modName ~= "", "modName", modName)
        and THUtils.argIsValid(THUtils.getIsType(target, THAbstractSystem), "target", target)
    then
        local moduleId = name:upper()
        local constId = THUtils.snakeCase(name, true)
        local modData = thModManager:getLoadedMod(modName)
        if modData == nil then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "modName", modName)
        elseif modData.name == self.modName then
            THUtils.stackMsg("%s is the main environment and cannot be a module", modData.name)
        elseif self.modules.byId[moduleId] ~= nil then
            THUtils.warningMsg("Module (%s) already exists", name)
        elseif self.MODULE[constId] ~= nil then
            THUtils.warningMsg("Module constant (%s) already exists", constId)
        else
            local moduleData = {
                id = moduleId,
                name = name,
                index = #self.modules.byIndex + 1,
                modData = modData,
                target = target,
                owner = self
            }
            self.modules.byId[moduleData.id] = moduleData
            self.modules.byIndex[moduleData.index] = moduleData
            self.MODULE[constId] = moduleData.index
            self:addEventListener(moduleData.target)
            return moduleData
        end
    end
end
function THAdvancedAddon:getModule(target)
    local targetType = type(target)
    local moduleData = nil -- placeholder
    if targetType == "string" then
        moduleData = self.modules.byId[target:upper()]
    elseif targetType == "number" then
        moduleData = self.modules.byIndex[target]
    elseif targetType == "table" and target.id ~= nil then
        moduleData = self.modules.byId[target.id]
        if moduleData ~= nil and moduleData ~= target then
            moduleData = nil
        end
    elseif target ~= nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "target", target)
    end
    if moduleData ~= nil then
        return moduleData.target, moduleData
    end
end
function THAdvancedAddon.hook_loadProfiles(superFunc, self, gui, xmlFilename, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local thModManager = self.thModManager
            for _, profile in pairs(gui.profiles) do
                if profile.values ~= nil then
                    for name, value in pairs(profile.values) do
                        if string.find(name, "Filename$") then
                            if type(value) == "string" then
                                local _, ePos, modName = string.find(value, "^%$(.+)%$")
                                if modName ~= nil then
                                    local modData = thModManager:getLoadedMod(modName)
                                    local filename = string.sub(value, ePos + 1)
                                    if modData ~= nil and filename ~= nil then
                                        local newValue = modData.path .. filename
                                        profile.values[name] = newValue
                                    end
                                end
                            end
                        end
                    end
                end
            end
        end)
        return ...
    end
    return appendFunc(superFunc(gui, xmlFilename, ...))
end
THUtils.call(function()
    local self = THAdvancedAddon.new()
    if self ~= nil then
        _G.g_thAdvancedAddon = self
        g_xmlManager:addCreateSchemaFunction(function()
            self.xmlSchema = XMLSchema.new(self.modName)
            self.xmlSchemaSettings = XMLSchema.new(self.modName .. "_settings")
            self.xmlSchemaSavegame = XMLSchema.new(self.modName .. "_savegame")
        end)
        g_xmlManager:addInitSchemaFunction(function()
            THAdvancedAddon.initXMLSchema(self.xmlSchema, self.xmlBaseKey)
            THAdvancedAddon.initXMLSchemaSettings(self.xmlSchemaSettings, self.xmlBaseKey)
            THAdvancedAddon.initXMLSchemaSavegame(self.xmlSchemaSavegame, self.xmlBaseKey)
        end)
        THUtils.setFunctionHook("Gui", "loadProfiles", THAdvancedAddon, nil, false, self)
        self.gui:loadProfiles(self.modPath .. "gui/guiProfiles.xml")
    end
end)