-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THFruitTypeManager = {}
local THFruitTypeManager_mt = THUtils.createClass(THFruitTypeManager, THAbstractSystem)
THFruitTypeManager.DATA_KEY = "fruitTypes"
local debugFlagId = THDebugUtil.createDebugFlag("THFruitTypeManager")
THFruitTypeManager.debugFlagId = debugFlagId
function THFruitTypeManager.new(owner, dataKey, modName, customMt)
    dataKey = dataKey or THFruitTypeManager.DATA_KEY
    customMt = customMt or THFruitTypeManager_mt
    local self = THAbstractSystem.new(owner, dataKey, modName, customMt)
    if self ~= nil then
        if owner.fruitTypeManager == nil then
            THUtils.stackMsg("Cannot find parent fruitTypeManager")
        else
            self.parent = owner.fruitTypeManager
            self.fruitTypes = {}
            self.fillTypeToFruitType = {}
            self.thEventManager:addEventListener(self)
            return self
        end
    end
end
function THFruitTypeManager:initializeDataTables()
    local fruitTypesArray, numFruitTypes = self:getFruitTypeList(true)
    if numFruitTypes > 0 then
        for fruitTypeIndex, fruitTypeInfo in ipairs(fruitTypesArray) do
            if self.fruitTypes[fruitTypeIndex] == nil then
                local fruitTypeData = {}
                local fruitFillTypeInfo = fruitTypeInfo.fillType
                local startStateChannel = tonumber(fruitTypeInfo.startStateChannel) or -1
                local numStateChannels = tonumber(fruitTypeInfo.numStateChannels) or 0
                local numGrowthStates = 0
                if startStateChannel >= 0 and numStateChannels > 0 then
                    numGrowthStates = (2 ^ (numStateChannels - startStateChannel)) - 1
                end
                if numGrowthStates <= 0 then
                    THUtils.warningMsg("FruitType (%s) does not have any registered growth states", fruitTypeInfo.name)
                    numGrowthStates = 0
                end
                fruitTypeData.index = fruitTypeInfo.index
                fruitTypeData.name = fruitTypeInfo.name
                fruitTypeData.info = fruitTypeInfo
                fruitTypeData.fillTypeIndex = fruitFillTypeInfo.index
                fruitTypeData.fillTypeName = fruitFillTypeInfo.name
                fruitTypeData.fillTypeInfo = fruitFillTypeInfo
                if fruitTypeInfo.hasWindrow and fruitTypeInfo.windrowName ~= nil then
                    local fruitWindrowInfo = self.thFillTypeManager:getFillType(fruitTypeInfo.windrowName)
                    if fruitWindrowInfo ~= nil and fruitWindrowInfo.index ~= FillType.UNKNOWN then
                        local fillTypeYield = tonumber(fruitTypeInfo.literPerSqm) or -1
                        local windrowYield = tonumber(fruitTypeInfo.windrowLiterPerSqm) or -1
                        fruitTypeData.windrowIndex = fruitWindrowInfo.index
                        fruitTypeData.windrowName = fruitWindrowInfo.name
                        fruitTypeData.windrowInfo = fruitWindrowInfo
                        fruitTypeData.windrowFactor = 1
                        if windrowYield > 0 then
                            if fillTypeYield > 0 then
                                fruitTypeData.windrowFactor = windrowYield / fillTypeYield
                            else
                                fruitTypeData.windrowFactor = 0
                            end
                        end
                    end
                end
                fruitTypeData.minHarvestState = tonumber(fruitTypeInfo.minHarvestingGrowthState) or -1
                fruitTypeData.maxHarvestState = tonumber(fruitTypeInfo.maxHarvestingGrowthState) or fruitTypeData.minHarvestState
                fruitTypeData.minForageState = tonumber(fruitTypeInfo.minForageGrowthState) or -1
                fruitTypeData.numGrowthStates = numGrowthStates
                if fruitTypeData.minHarvestState > 0 and fruitTypeData.minForageState > 0 then
                    fruitTypeData.maxForageState = math.max(fruitTypeData.minForageState, fruitTypeData.minHarvestState - 1)
                else
                    fruitTypeData.maxForageState = fruitTypeData.minForageState
                end
                self.fruitTypes[fruitTypeInfo.index] = fruitTypeData
            end
        end
    end
    return true
end
function THFruitTypeManager:finalizeDataTables()
    local fruitTypesArray, numFruitTypes = self:getFruitTypeList(true)
    THUtils.clearTable(self.fillTypeToFruitType)
    if numFruitTypes > 0 then
        for fruitTypeIndex, fruitTypeInfo in ipairs(fruitTypesArray) do
            local fruitTypeData = self:getFruitTypeData(fruitTypeIndex)
            if fruitTypeData == nil then
                THUtils.stackMsg(THMessage.INITIALIZATION_FAILED, "fruitType", fruitTypeInfo.name)
            else
                local fillToFruitTypeData = self.fillTypeToFruitType[fruitTypeData.fillTypeIndex]
                if fillToFruitTypeData ~= nil then
                    if fillToFruitTypeData.fillTypeIndex ~= fruitTypeData then
                        THUtils.warningMsg("FillType (%s) already registered to fruitType (%s)", fruitTypeData.fillTypeName, fillToFruitTypeData.name)
                    end
                else
                    self.fillTypeToFruitType[fruitTypeData.fillTypeIndex] = fruitTypeData
                end
            end
        end
    end
    return true
end
function THFruitTypeManager:onPostInit()
    if not self.isPostInitFinished then
        THFruitTypeManager:superClass().onPostInit(self)
        self.thFillTypeManager = self.mainEnvironment.thFillTypeManager
        self:addDelayedLoadTask(function()
            THUtils.setFunctionHook("FruitTypeManager", "loadMapData", THFruitTypeManager, nil, false, self)
            THUtils.setFunctionHook("GrowthSystem", "loadMapData", THFruitTypeManager, "hook_loadGrowthSystemData", false, self)
            THUtils.setFunctionHook("MotionPathEffectManager", "loadMapData", THFruitTypeManager, "hook_loadMotionPathEffectData", false, self)
        end)
        self.isPostInitFinished = true
    end
end
function THFruitTypeManager:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
    THFruitTypeManager:superClass().onSetMissionInfo(self, mission, missionInfo, missionDynamicInfo)
end
function THFruitTypeManager:onLoadMapFinished(mission)
    THFruitTypeManager:superClass().onLoadMapFinished(self, mission)
    self:initializeDataTables()
    self:finalizeDataTables()
end
function THFruitTypeManager:getFruitType(fruitType)
    local manager = self.parent
    local targetType = type(fruitType)
    local fruitTypeInfo = nil
    if targetType == "number" then
        fruitTypeInfo = manager:getFruitTypeByIndex(fruitType)
    elseif targetType == "string" then
        fruitTypeInfo = manager:getFruitTypeByName(fruitType:upper())
    elseif targetType == "table" and fruitType.name ~= nil then
        fruitTypeInfo = manager:getFruitTypeByName(fruitType.name)
        if fruitTypeInfo ~= fruitType then
            fruitTypeInfo = nil
        end
    elseif fruitType ~= nil then -- wrong variable type
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "fruitType", fruitType, true)
    end
    return fruitTypeInfo
end
function THFruitTypeManager:getFruitTypeName(fruitType)
    local fruitTypeInfo = self:getFruitType(fruitType)
    if fruitTypeInfo ~= nil and fruitTypeInfo.name ~= nil then
        return fruitTypeInfo.name
    end
    return "UNKNOWN"
end
function THFruitTypeManager:getFruitTypeByFillType(fillType, ...)
    local fillTypeInfo = self.thFillTypeManager:getFillType(fillType)
    if fillTypeInfo ~= nil then
        return self.parent:getFruitTypeByFillTypeIndex(fillTypeInfo.index, ...)
    end
end
function THFruitTypeManager:getFruitTypeData(fruitType)
    local fruitTypeInfo = self:getFruitType(fruitType)
    if fruitTypeInfo ~= nil then
        return self.fruitTypes[fruitTypeInfo.index]
    end
end
function THFruitTypeManager:getFruitTypeDataByFillType(fillType)
    local fillTypeInfo = self.thFillTypeManager:getFillType(fillType)
    if fillTypeInfo ~= nil then
        return self.fillTypeToFruitType[fillTypeInfo.index]
    end
end
function THFruitTypeManager:getFruitTypeList(byIndex)
    local manager = self.parent
    local fruitTypesArray = manager:getFruitTypes()
    local numFruitTypes = 0
    if fruitTypesArray ~= nil then
        numFruitTypes = #fruitTypesArray
    end
    if byIndex == true then
        return fruitTypesArray, numFruitTypes
    end
    return manager.nameToFruitType, numFruitTypes
end
function THFruitTypeManager:getFruitTypeFillType(fruitType)
    local fruitTypeData = self:getFruitTypeData(fruitType)
    local fillTypeInfo = nil
    if fruitTypeData ~= nil then
        fillTypeInfo = fruitTypeData.fillTypeInfo
    end
    if fillTypeInfo == nil then
        local fruitTypeInfo = self:getFruitType(fruitType)
        if fruitTypeInfo ~= nil then
            if fruitTypeInfo.fillType ~= nil then
                fillTypeInfo = fruitTypeInfo.fillType
            else
                fillTypeInfo = self.parent:getFillTypeByFruitTypeIndex(fruitTypeInfo.index)
            end
        end
    end
    return fillTypeInfo
end
function THFruitTypeManager:getFruitTypeWindrowFillType(fruitType)
    local fruitTypeData = self:getFruitTypeData(fruitType)
    local windrowFillTypeInfo = nil
    if fruitTypeData ~= nil then
        windrowFillTypeInfo = fruitTypeData.windrowInfo
    end
    if windrowFillTypeInfo == nil then
        local fruitTypeInfo = self:getFruitType(fruitType)
        if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= FruitType.UNKNOWN
            and fruitTypeInfo.hasWindrow
        then
            if fruitTypeInfo.windrowName ~= nil then
                windrowFillTypeInfo = self.thFillTypeManager:getFillType(fruitTypeInfo.windrowName)
            end
            if windrowFillTypeInfo == nil then
                local windrowIndex = self.parent:getWindrowFillTypeIndexByFruitTypeIndex(fruitTypeInfo.index)
                windrowFillTypeInfo = self.thFillTypeManager:getFillType(windrowIndex)
            end
        end
    end
    return windrowFillTypeInfo
end
function THFruitTypeManager:getFruitTypeYield(fruitType, fillType, defaultValue)
    defaultValue = defaultValue or 0
    if THUtils.argIsValid(type(defaultValue) == "number" and defaultValue >= 0, "defaultValue", defaultValue) then
        local fruitTypeInfo = self:getFruitType(fruitType)
        if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= FruitType.UNKNOWN then
            local literPerSqm = tonumber(fruitTypeInfo.literPerSqm) or -1
            local fillTypeIndex = FillType.UNKNOWN
            if fillType ~= nil then
                local fillTypeInfo = self.thFillTypeManager:getFillType(fillType)
                if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
                    fillTypeIndex = fillTypeInfo.index
                end
            end
            if fillTypeIndex ~= FillType.UNKNOWN then
                if fruitTypeInfo.hasWindrow and fruitTypeInfo.windrowName ~= nil then
                    local windrowLiterPerSqm = tonumber(fruitTypeInfo.windrowLiterPerSqm) or -1
                    if windrowLiterPerSqm >= 0 then
                        local windrowInfo = self:getFruitTypeWindrowFillType(fruitTypeInfo.index)
                        if windrowInfo ~= nil and windrowInfo.index ~= FillType.UNKNOWN
                            and windrowInfo.index == fillTypeIndex
                        then
                            return windrowLiterPerSqm
                        end
                    end
                end
            end
            if literPerSqm >= 0 then
                if fillType == nil
                    or (fillTypeIndex ~= FillType.UNKNOWN
                        and fruitTypeInfo.fillType ~= nil
                        and fruitTypeInfo.fillType.index == fillTypeIndex)
                then
                    return literPerSqm
                end
            end
        end
        return defaultValue
    end
    return 0
end
function THFruitTypeManager:getFruitTypeConverterData(converter, ...)
    local manager = self.parent
    if type(converter) == "string" then
        return manager:getConverterDataByName(converter, ...)
    elseif converter ~= nil then
        if manager.fruitTypeConverters ~= nil then
            return manager.fruitTypeConverters[converter]
        end
    end
end
function THFruitTypeManager:addFruitTypeConversion(name, isBaseType, fruitType, fillType, factor, windrowFactor, ...)
    local manager = self.parent
    local fruitTypeInfo = self:getFruitType(fruitType)
    local fillTypeInfo = self.thFillTypeManager:getFillType(fillType)
    if fruitTypeInfo ~= nil and fillTypeInfo ~= nil then
        if type(name) == "string" then
            local converterId = name:upper()
            local converterIndex = manager.converterNameToIndex[converterId]
            if converterIndex == nil then
                converterIndex = manager:addFruitTypeConverter(name, isBaseType)
                if converterIndex == nil then
                    return
                end
            end
            manager:addFruitTypeConversion(converterIndex, fruitTypeInfo.index, fillTypeInfo.index, factor, windrowFactor, ...)
            local converterData = manager.fruitTypeConverters[converterIndex]
            if converterData ~= nil then
                return converterData, converterData[fruitTypeInfo.index]
            end
        end
    end
end
function THFruitTypeManager:getFruitTypeGrowthState(fruitType, stateId, isStrict)
    local fruitTypeData = self:getFruitTypeData(fruitType)
    local stateIdType = type(stateId)
    local growthState = -1
    local growthStateName = "UNKNOWN"
    if fruitTypeData ~= nil and stateId ~= nil then
        local fruitTypeInfo = fruitTypeData.info
        local startStateChannel = fruitTypeInfo.startStateChannel
        local numStateChannels = fruitTypeInfo.numStateChannels
        local maxNumStates = 0
        if startStateChannel ~= nil and numStateChannels ~= nil then
            maxNumStates = (2 ^ (numStateChannels - startStateChannel)) - 1
        end
        if maxNumStates > 0 then
            local numGrowthStates = THUtils.getNoNil(fruitTypeData.numGrowthStates, 0)
            local minHarvestState = THUtils.getNoNil(fruitTypeData.minHarvestState, -1)
            local maxHarvestState = THUtils.getNoNil(fruitTypeData.maxHarvestState, minHarvestState)
            local minForageState = THUtils.getNoNil(fruitTypeData.minForageState, -1)
            local maxForageState = THUtils.getNoNil(fruitTypeData.maxForageState, minForageState)
            local minPreparingState = THUtils.getNoNil(fruitTypeInfo.minPreparingGrowthState, -1)
            local maxPreparingState = THUtils.getNoNil(fruitTypeInfo.maxPreparingGrowthState, minPreparingState)
            local preparedState = THUtils.getNoNil(fruitTypeInfo.preparedGrowthState, -1)
            local witheredState = THUtils.getNoNil(fruitTypeInfo.witheredState, -1)
            local cutState = THUtils.getNoNil(fruitTypeInfo.cutState, -1)
            if stateIdType == "string" then
                growthState = tonumber(stateId)
            elseif stateIdType == "number" then
                growthState = stateId
            end
            if growthState ~= nil then
                growthState = math.floor(growthState)
                if growthState < 0 or growthState > maxNumStates then
                    if isStrict == true then
                        growthState = -1
                    else
                        growthState = MathUtil.clamp(growthState, 0, maxNumStates)
                    end
                end
            elseif stateIdType == "string" then
                local _, e1, stateName = string.find(stateId, "^(%w+)")
                local stateOffset1 = nil
                local stateOffset2 = nil
                if e1 ~= nil and e1 > 0 and stateName ~= nil then
                    local _, e2, offset1 = string.find(stateId, "_(%w+)", e1 + 1)
                    if e2 ~= nil and e2 > 0 and offset1 ~= nil then
                        stateOffset1 = offset1 -- string (i.e. MIN, MAX)
                    else
                        e2 = e1
                    end
                    local _, _, offset2 = string.find(stateId, "([+-]%d+)", e2 + 1)
                    if offset2 ~= nil then
                        stateOffset2 = tonumber(offset2) -- value (i.e. -1, 1)
                    end
                end
                if stateName ~= nil then
                    local minState, maxState = -1, -1
                    stateName = stateName:upper()
                    if stateName == "GROWTH" then
                        if numGrowthStates > 0 then
                            minState = 1
                            maxState = numGrowthStates
                        end
                    elseif stateName == "FORAGE" then
                        minState = minForageState
                        maxState = maxForageState
                    elseif stateName == "PREPARING" then
                        minState = minPreparingState
                        maxState = maxPreparingState
                    elseif stateName == "PREPARED" then
                        minState = preparedState
                        maxState = preparedState
                    elseif stateName == "HARVEST" then
                        minState = minHarvestState
                        maxState = maxHarvestState
                    elseif stateName == "WITHERED" then
                        minState = witheredState
                        maxState = witheredState
                    elseif stateName == "CUT" then
                        minState = cutState
                        maxState = maxNumStates
                    elseif stateName == "EXTRA" then
                        if cutState > 0 and cutState + 2 <= maxNumStates then
                            minState = cutState + 2
                            maxState = maxNumStates
                        end
                    end
                    if minState > 0 and maxState > 0 then
                        if stateOffset1 ~= nil then
                            stateOffset1 = stateOffset1:upper()
                        end
                        if stateOffset1 == nil or stateOffset1 == "MIN" then
                            stateOffset1 = 0
                        elseif stateOffset1 == "MAX" then
                            stateOffset1 = maxState - minState
                        else
                            stateOffset1 = tonumber(stateOffset1)
                            stateOffset1 = THUtils.getNoNil(stateOffset1, 0) - 1
                        end
                        if stateOffset1 >= 0 then
                            growthState = math.min(minState + stateOffset1, maxState)
                            if stateOffset2 ~= nil then
                                growthState = growthState + stateOffset2
                                if growthState < 0 or growthState > maxNumStates then
                                    if isStrict == true then
                                        growthState = -1
                                    else
                                        growthState = MathUtil.clamp(growthState, 0, maxNumStates)
                                    end
                                end
                            end
                            if growthState >= 0 then
                                growthStateName = stateName
                            end
                        end
                    end
                end
            end
        end
    end
    return growthState, growthStateName
end
function THFruitTypeManager.hook_loadMapData(superFunc, self, manager, xmlFile, missionInfo, baseDirectory, ...)
    local function appendFunc(rSuccess, ...)
        if rSuccess and self.mapSettingsModData ~= nil then
            THUtils.call(function()
                self:loadMapSettings("fruitTypes", nil, nil, manager.loadFruitTypes, manager, missionInfo, false)
                THDebugUtil.printTable(debugFlagId, nil, manager.fruitTypes, 1)
            end)
        end
        return rSuccess, ...
    end
    return appendFunc(superFunc(manager, xmlFile, missionInfo, baseDirectory, ...))
end
function THFruitTypeManager.hook_loadGrowthSystemData(superFunc, self, growthSystem, xmlFile, missionInfo, baseDirectory, ...)
    local function appendFunc(...)
        THUtils.call(function()
            if self.mapSettingsModData ~= nil then
                self:loadMapSettings("growth", true, nil, growthSystem.loadGrowthData, growthSystem, "growth")
            end
            local numGrowthPeriods = Environment.PERIODS_IN_YEAR or 0
            if numGrowthPeriods > 0 then
                local function updateGrowthMapping(pDataTable)
                    if pDataTable ~= nil then
                        for fruitTypeIndex, growthInfo in pairs(pDataTable) do
                            local fruitTypeInfo = self:getFruitType(fruitTypeIndex)
                            if fruitTypeInfo ~= nil then
                                local numStateChannels = fruitTypeInfo.numStateChannels or 0
                                local maxNumStates = (2 ^ numStateChannels - 1)
                                if numStateChannels > 0 and maxNumStates > 0 then
                                    for periodIndex = 1, numGrowthPeriods do
                                        local periodData = growthInfo[periodIndex]
                                        if periodData ~= nil and periodData.growthMapping ~= nil
                                            and next(periodData.growthMapping) ~= nil
                                        then
                                            for growthState = 1, maxNumStates do
                                                if periodData.growthMapping[growthState] == nil then
                                                    periodData.growthMapping[growthState] = growthState
                                                end
                                            end
                                        end
                                    end
                                end
                            end
                        end
                    end
                end
                updateGrowthMapping(growthSystem.seasonalFruitData)
                updateGrowthMapping(growthSystem.nonSeasonalFruitData)
            end
        end)
        return ...
    end
    return appendFunc(superFunc(growthSystem, xmlFile, missionInfo, baseDirectory, ...))
end
function THFruitTypeManager.hook_loadMotionPathEffectData(superFunc, self, manager, xmlFile, missionInfo, baseDirectory, ...)
    local function appendFunc(rSuccess, ...)
        if rSuccess and self.mapSettingsModData ~= nil then
            THUtils.call(function()
                self:loadMapSettings("motionPathEffects", nil, nil, manager.loadMotionPathEffects, manager, "motionPathEffects.motionPathEffect", self.mapSettingsModPath, self.mapSettingsModName)
            end)
        end
        return rSuccess, ...
    end
    return appendFunc(superFunc(manager, xmlFile, missionInfo, baseDirectory, ...))
end