-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THPrecisionFarmingManager = {}
local THPrecisionFarmingManager_mt = THUtils.createClass(THPrecisionFarmingManager, THAbstractSystem)
THPrecisionFarmingManager.DATA_KEY = "precisionFarming"
local debugFlagId = THDebugUtil.createDebugFlag("THPrecisionFarming")
THPrecisionFarmingManager.debugFlagId = debugFlagId
function THPrecisionFarmingManager.new(owner, dataKey, modName, customMt)
    dataKey = dataKey or THPrecisionFarmingManager.DATA_KEY
    customMt = customMt or THPrecisionFarmingManager_mt
    local self = THAbstractSystem.new(owner, dataKey, modName, customMt)
    self.thModManager = owner.thModManager
    self.thSettingsManager = owner.thSettingsManager
    self.minSoilTypeYieldPotential = 0
    self.maxSoilTypeYieldPotential = 1
    self.functionParameters = {
        forceIsUncovered = false
    }
    self.thEventManager:addEventListener(self)
    self.isEnabled = false
    return self
end
function THPrecisionFarmingManager:initDataTables()
    if self:initSoilMapData() then
        return true
    end
    return false
end
function THPrecisionFarmingManager:initSoilMapData()
    local pfMain = self.parent
    if pfMain ~= nil then
        local soilTypesArray, numSoilTypes = self:getSoilTypeList()
        if soilTypesArray ~= nil and numSoilTypes > 0 then
            local minYieldPotential = math.huge
            local maxYieldPotential = -math.huge
            for soilTypeIndex = 1, numSoilTypes do
                local soilTypeInfo = soilTypesArray[soilTypeIndex]
                local yieldPotential = soilTypeInfo.yieldPotential or 1
                minYieldPotential = math.min(minYieldPotential, yieldPotential)
                maxYieldPotential = math.max(maxYieldPotential, yieldPotential)
            end
            if minYieldPotential ~= math.huge and maxYieldPotential ~= -math.huge then
                self.minSoilTypeYieldPotential = math.min(minYieldPotential, maxYieldPotential)
                self.maxSoilTypeYieldPotential = math.max(minYieldPotential, maxYieldPotential)
            else
                self.minSoilTypeYieldPotential = 0
                self.maxSoilTypeYieldPotential = 1
            end
        end
    end
    return true
end
function THPrecisionFarmingManager:onPreInit()
    if not self.isPreInitFinished then
        THPrecisionFarmingManager:superClass().onPreInit(self)
        local thModManager = self.thModManager
        local pfModData = thModManager:getLoadedMod("FS22_precisionFarming")
        if pfModData == nil then
            pfModData = thModManager:getLoadedMod("FS22_PrecisionFarmingAnhydrousReady")
        end
        if pfModData ~= nil then
            local pfModEnv = THUtils.getModEnvironment(pfModData.name)
            self.parent = pfModEnv.g_precisionFarming
            self.pfSoilMap = self.parent.soilMap
            self.pfModData = pfModData
            self.pfModEnvironment = pfModEnv
            self.isEnabled = true
            local CoverMap = pfModEnv.CoverMap
            THUtils.setFunctionHook(CoverMap, "getIsUncoveredAtBitVectorPos", THPrecisionFarmingManager.CoverMap, nil, true, self.functionParameters)
        end
        self.isPreInitFinished = true
    end
end
function THPrecisionFarmingManager:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
    THPrecisionFarmingManager:superClass().onSetMissionInfo(self, mission, missionInfo, missionDynamicInfo)
    self:preLoadConfiguration()
end
function THPrecisionFarmingManager:onLoadMapFinished(mission)
    THPrecisionFarmingManager:superClass().onLoadMapFinished(self, mission)
    self:initDataTables()
    self:loadConfiguration()
end
function THPrecisionFarmingManager:getSoilTypeIndexAtWorldPos(x, z, force)
    local pfSoilMap = self.pfSoilMap
    local soilTypeIndex = 0
    if pfSoilMap ~= nil then
        self.functionParameters.forceIsUncovered = force == true
        soilTypeIndex = pfSoilMap:getTypeIndexAtWorldPos(x, z)
        self.functionParameters.forceIsUncovered = false
    end
    if soilTypeIndex ~= nil and soilTypeIndex > 0 then
        return soilTypeIndex
    end
    return 0
end
function THPrecisionFarmingManager:getSoilTypeAtWorldPos(x, z, force)
    local soilTypeIndex = self:getSoilTypeIndexAtWorldPos(x, z, force)
    if soilTypeIndex > 0 then
        return self:getSoilType(soilTypeIndex)
    end
end
function THPrecisionFarmingManager:getSoilType(soilTypeIndex)
    local pfSoilMap = self.pfSoilMap
    if pfSoilMap ~= nil and soilTypeIndex ~= nil and soilTypeIndex > 0 then
        return pfSoilMap:getSoilTypeByIndex(soilTypeIndex)
    end
end
function THPrecisionFarmingManager:getSoilTypeList()
    local pfSoilMap = self.pfSoilMap
    if pfSoilMap ~= nil and pfSoilMap.soilTypes ~= nil then
        return pfSoilMap.soilTypes, #pfSoilMap.soilTypes
    end
    return nil, 0
end
THPrecisionFarmingManager.CoverMap = {}
function THPrecisionFarmingManager.CoverMap:hook_getIsUncoveredAtBitVectorPos(superFunc, extraArgs, ...)
    local function appendFunc(rIsUncovered, ...)
        if extraArgs.forceIsUncovered == true then
            rIsUncovered = true
        end
        return rIsUncovered, ...
    end
    return appendFunc(superFunc(self, ...))
end